@@ -11,11 +11,19 @@ import pycountry
 import validators
 import tzlocal
 import getpass
-from zoneinfo import available_timezones
+import datetime
+import ssl
+import uuid
 from aiohttp import web
+from zoneinfo import available_timezones
 from passlib.hash import sha512_crypt
 from pathlib import Path
 from . import __version__
+from cryptography import x509
+from cryptography.x509.oid import NameOID
+from cryptography.hazmat.primitives import serialization
+from cryptography.hazmat.primitives.asymmetric import ed25519
+from cryptography.hazmat.primitives import hashes
 
 valid_keyboard_layouts = [
     "de",
@@ -49,6 +57,12 @@ indie_root_dir = Path.cwd() / ".indie"
 indie_toml_file = indie_root_dir / "indie.toml"
 indie_toml = tomlkit.document()
 
+private_key_pem_file = indie_root_dir / "private_key.pem"
+cert_pem_file = indie_root_dir / "cert.pem"
+
+private_key_ssh_file = indie_root_dir / "id_ed25519"
+public_key_ssh_file = indie_root_dir / "id_ed25519.pub"
+
 
 def write_toml(data: dict):
     indie_toml.update(data)
@@ -65,10 +79,12 @@ def get_proxmox_toml(mac):
             continue
 
         domain = toml_dict["global"].pop("domain")
-        host_dict = {"global": toml_dict["global"]}
-        host_dict["global"]["fqdn"] = host["hostname"] + "." + domain
+        token = toml_dict["global"].pop("https-access-token")
+        fingerprint = toml_dict["global"].pop("cert-fingerprint")
+        to_write = {"global": toml_dict["global"]}
+        to_write["global"]["fqdn"] = host["hostname"] + "." + domain
 
-        host_dict["network"] = (
+        to_write["network"] = (
             {"source": "from-dhcp"}
             if host["use-dhcp"]
             else {
@@ -82,28 +98,30 @@ def get_proxmox_toml(mac):
 
         # TODO: Improve on this
         if host["use-raid1"]:
-            host_dict["disk-setup"] = {
+            to_write["disk-setup"] = {
                 "disk-list": ["sda", "sdb", "sdc", "sdd", "sde", "sdf"],
                 "filesystem": "zfs",
                 tomlkit.key(["zfs", "raid"]): "raid1",
             }
         else:
-            host_dict["disk-setup"] = {
+            to_write["disk-setup"] = {
                 "disk-list": ["sda"],
                 "filesystem": "ext4",
             }
 
-        host_dict["first-boot"] = {
+        to_write["first-boot"] = {
             "source": "from-url",
-            "url": f"http://indie.{domain}:8000/proxmox-first-boot",
+            "url": f"https://indie.{domain}:8000/proxmox-first-boot?token={token}",
+            "cert-fingerprint": fingerprint,
         }
 
-        host_dict["post-installation-webhook"] = {
-            "url": f"http://indie.{domain}:8000/proxmox-post-install",
+        to_write["post-installation-webhook"] = {
+            "url": f"https://indie.{domain}:8000/proxmox-post-install?token={token}",
+            "cert-fingerprint": fingerprint,
         }
 
         proxmox_toml = tomlkit.document()
-        proxmox_toml.update(host_dict)
+        proxmox_toml.update(to_write)
         return proxmox_toml
     return None
 
@@ -115,6 +133,85 @@ def get_toml_default(key):
         return None
 
 
+def generate_cert_and_ssh_keys_if_not_present():
+    if private_key_pem_file.is_file() and cert_pem_file.is_file():
+        with open(private_key_pem_file, "rb") as file:
+            private_key = serialization.load_pem_private_key(file.read(), None)
+        with open(cert_pem_file, "rb") as file:
+            cert = x509.load_pem_x509_certificate(file.read())
+    else:
+        private_key = ed25519.Ed25519PrivateKey.generate()
+        subject = issuer = x509.Name(
+            [
+                x509.NameAttribute(NameOID.COUNTRY_NAME, get_toml_default("country")),
+                x509.NameAttribute(NameOID.EMAIL_ADDRESS, get_toml_default("mailto")),
+                x509.NameAttribute(
+                    NameOID.LOCALITY_NAME,
+                    get_toml_default("timezone").split("/")[1].replace("_", " "),
+                ),
+                x509.NameAttribute(NameOID.COMMON_NAME, get_toml_default("domain")),
+            ]
+        )
+        cert = (
+            x509.CertificateBuilder()
+            .subject_name(subject)
+            .issuer_name(issuer)
+            .public_key(private_key.public_key())
+            .serial_number(x509.random_serial_number())
+            .not_valid_before(datetime.datetime.now(datetime.timezone.utc))
+            .not_valid_after(
+                # Our certificate will be valid for 100 years
+                datetime.datetime.now(datetime.timezone.utc)
+                + datetime.timedelta(days=365 * 100)
+            )
+            .add_extension(
+                x509.SubjectAlternativeName([x509.DNSName(get_toml_default("domain"))]),
+                critical=False,
+            )
+            .sign(private_key, None)
+        )
+
+        private_key_pem_file.parent.mkdir(parents=True, exist_ok=True)
+        with open(private_key_pem_file, "wb") as file:
+            file.write(
+                private_key.private_bytes(
+                    encoding=serialization.Encoding.PEM,
+                    format=serialization.PrivateFormat.PKCS8,
+                    encryption_algorithm=serialization.NoEncryption(),
+                )
+            )
+
+        cert_pem_file.parent.mkdir(parents=True, exist_ok=True)
+        with open(cert_pem_file, "wb") as file:
+            file.write(cert.public_bytes(serialization.Encoding.PEM))
+
+    private_key_ssh_file.parent.mkdir(parents=True, exist_ok=True)
+    with open(private_key_ssh_file, "wb") as file:
+        file.write(
+            private_key.private_bytes(
+                encoding=serialization.Encoding.PEM,
+                format=serialization.PrivateFormat.OpenSSH,
+                encryption_algorithm=serialization.NoEncryption(),
+            )
+        )
+
+    public_key_ssh_file.parent.mkdir(parents=True, exist_ok=True)
+    with open(public_key_ssh_file, "wb") as file:
+        file.write(
+            private_key.public_key().public_bytes(
+                encoding=serialization.Encoding.OpenSSH,
+                format=serialization.PublicFormat.OpenSSH,
+            )
+        )
+
+    fingerprint = cert.fingerprint(hashes.SHA256())
+    to_write = indie_toml.unwrap()
+    to_write["global"]["cert-fingerprint"] = fingerprint.hex(":")
+    with open(public_key_ssh_file, "r") as file:
+        to_write["global"]["root-ssh-keys"] = [file.read()]
+    write_toml(to_write)
+
+
 def get_keyboard(args):
     selected_keyboard_layout = args.keyboard
     while selected_keyboard_layout not in valid_keyboard_layouts:
@@ -169,7 +266,7 @@ def get_domain(args):
 
         selected_domain = input()
     print(f"Selected domain: {selected_domain}")
-    return selected_domain
+    return selected_domain.lower()
 
 
 def get_mailto(args):
@@ -235,10 +332,20 @@ def get_password(args):
         if not sha512_crypt.verify(getpass.getpass(), selected_password):
             print("Entered passwords didn't match, try again")
             selected_password = None
+            continue
+
     print(f"Selected password hash: {selected_password}")
     return selected_password
 
 
+def get_https_access_token(args):
+    selected_https_access_token = args.https_access_token
+    while not validators.uuid(selected_https_access_token):
+        selected_https_access_token = str(uuid.uuid4())
+    print(f"Selected https_access_token: {selected_https_access_token}")
+    return selected_https_access_token.lower()
+
+
 def command_begin(args):
     begin_dict = {
         "domain": get_domain(args),
@@ -247,10 +354,12 @@ def command_begin(args):
         "country": get_country(args),
         "timezone": get_timezone(args),
         "root-password-hashed": get_password(args),
+        "https-access-token": get_https_access_token(args),
     }
     write_toml(
         {"global": begin_dict},
     )
+    generate_cert_and_ssh_keys_if_not_present()
 
 
 def is_hostproperty_in_use(prop, cmp):
@@ -285,10 +394,11 @@ def get_macaddress(args):
     while not validators.mac_address(selected_macaddress) or is_hostproperty_in_use(
         "macaddress", selected_macaddress
     ):
-        if not validators.mac_address(selected_macaddress):
-            print(f"{selected_macaddress} is not a valid MAC address")
-        elif is_hostproperty_in_use("macaddress", selected_macaddress):
-            print(f"'{selected_macaddress}' already in use")
+        if selected_macaddress is not None:
+            if not validators.mac_address(selected_macaddress):
+                print(f"{selected_macaddress} is not a valid MAC address")
+            elif is_hostproperty_in_use("macaddress", selected_macaddress):
+                print(f"'{selected_macaddress}' already in use")
         print("Select MAC address:")
 
         selected_macaddress = input()
@@ -370,6 +480,7 @@ def command_addhost(args):
         "country": get_country(args),
         "timezone": get_timezone(args),
         "root-password-hashed": get_password(args),
+        "https-access-token": get_https_access_token(args),
     }
 
     addhost_dict = {
@@ -398,13 +509,21 @@ def command_addhost(args):
 
     to_write.setdefault("host", []).append(addhost_dict)
     write_toml(to_write)
+    generate_cert_and_ssh_keys_if_not_present()
 
 
 def command_serve(args):
     routes = web.RouteTableDef()
+    token = get_toml_default("https-access-token")
+    print(f"Serving requests for {token}...")
 
     @routes.post("/proxmox-answer")
     async def proxmox_answer(request: web.Request):
+        if request.query.get("token", "") != token:
+            return web.Response(
+                status=401,
+                text=f"Unauthorized",
+            )
         try:
             request_data = json.loads(await request.text())
         except json.JSONDecodeError as e:
@@ -439,6 +558,11 @@ def command_serve(args):
 
     @routes.get("/proxmox-first-boot")
     async def proxmox_first_boot(request: web.Request):
+        if request.query.get("token", "") != token:
+            return web.Response(
+                status=401,
+                text=f"Unauthorized",
+            )
         print(f"Request proxmox-first-boot data for peer '{request.remote}':")
         domain = get_toml_default("domain")
         json_data = '"{\\"hostname\\":\\"$hostname\\",\\"message\\":\\"$1\\"}"'
@@ -447,7 +571,7 @@ set -ex
 
 hostname=$(hostname)
 report_progress() {{
-    curl --json {json_data} http://indie.{domain}:8000/report-progress
+    curl --insecure --json {json_data} https://indie.{domain}:8000/report-progress?token={token}
 }}
 report_progress "Re-writing apt sources..."
 sed -i 's|URIs: https://enterprise.proxmox.com/debian/ceph-squid|URIs: http://download.proxmox.com/debian/ceph-squid|g' /etc/apt/sources.list.d/ceph.sources
@@ -465,11 +589,15 @@ apt purge proxmox-first-boot
 report_progress "Script in first-boot completed successfully, server will now reboot a final time"
 reboot
 """
-        # print(message)
         return web.Response(text=message)
 
     @routes.post("/proxmox-post-install")
     async def proxmox_post_install(request: web.Request):
+        if request.query.get("token", "") != token:
+            return web.Response(
+                status=401,
+                text=f"Unauthorized",
+            )
         try:
             request_data = json.loads(await request.text())
         except json.JSONDecodeError as e:
@@ -488,6 +616,11 @@ reboot
 
     @routes.post("/report-progress")
     async def report_progress(request: web.Request):
+        if request.query.get("token", "") != token:
+            return web.Response(
+                status=401,
+                text=f"Unauthorized",
+            )
         try:
             request_data = json.loads(await request.text())
         except json.JSONDecodeError as e:
@@ -501,9 +634,15 @@ reboot
 
     @routes.get("/proxmox-create-usb-installer")
     async def proxmox_create_usb_installer(request: web.Request):
+        if request.query.get("token", "") != token:
+            return web.Response(
+                status=401,
+                text=f"Unauthorized",
+            )
         print(f"Request proxmox-create-usb-installer for peer '{request.remote}':")
         iso_name = "proxmox-ve_9.0-1"
         domain = get_toml_default("domain")
+        fingerprint = get_toml_default("cert-fingerprint")
         message = f"""#!/bin/bash
 set -ex
 if [[ $# -ne 1 ]]; then
@@ -519,7 +658,7 @@ fatlabel $1 "AUTOPROXMOX"
 
 # Get ISO
 wget https://enterprise.proxmox.com/iso/{iso_name}.iso
-proxmox-auto-install-assistant prepare-iso {iso_name}.iso --fetch-from http --url "http://indie.{domain}:8000/proxmox-answer" --output {iso_name}-auto.iso
+proxmox-auto-install-assistant prepare-iso {iso_name}.iso --fetch-from http --url "https://indie.{domain}:8000/proxmox-answer?token={token}" --cert-fingerprint "{fingerprint}" --output {iso_name}-auto.iso
 dd bs=1M conv=fdatasync if=./{iso_name}-auto.iso of=$1
 rm {iso_name}.iso
 rm {iso_name}-auto.iso
@@ -529,7 +668,9 @@ rm {iso_name}-auto.iso
 
     app = web.Application()
     app.add_routes(routes)
-    web.run_app(app, port=8000)
+    ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
+    ssl_context.load_cert_chain(cert_pem_file, private_key_pem_file)
+    web.run_app(app, port=8000, ssl_context=ssl_context)
 
 
 def command_unknown(args, parser):
@@ -572,6 +713,11 @@ def set_subparser_settings(subparser):
         default=get_toml_default("root-password-hashed"),
         help="SHA512 password hash, compatible with '/etc/shadow'",
     )
+    subparser.add_argument(
+        "--https-access-token",
+        default=get_toml_default("https-access-token"),
+        help="A UUID to limit access to the https webserver when running 'indie serve'",
+    )
 
 
 def main():
